/*
 [ GPL ]
 Code by JaeHyuk Cho <mailto:minzkn@infoeq.com>

 Release 1.0.0 - 2002.12.24 
 Release 1.0.1 - 2003.07.23
 
 - Simple is best !
*/

/* Normal include */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>
#include <time.h>
#include <signal.h>

/* Network include */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <net/if.h>

/* Debug 1=enable, 0=disable */
#define DEF_DEBUG_PROXY      (0) /* Release = 0 */
#define DEF_DEBUG_NODAEMON   (0) /* Release = 0 */

#define DEF_DEFAULT_BACKLOG  ( 5 )
#define DEF_DEFAULT_BUFFER   ( 128 << 10 )
#define DEF_DEFAULT_MAX      ( 100 )

#if DEF_DEBUG_NODAEMON == (0)
#define DEF_DEFAULT_LOG      "/var/log/mzproxy.log"
#else
#define DEF_DEFAULT_LOG      ((void *)0)
#endif

/* Proxy operation linked list */
typedef struct ts_MZ_Proxy
{ 
 struct ts_MZ_Proxy *Next;
 int                 ServerSocket, ClientSocket;
 struct sockaddr_in  AcceptAddress;
 socklen_t           AcceptAddressLength;
 unsigned long long  FromServerBytes, FromClientBytes;
}t_MZ_Proxy;

int main(int s_Argc, char *s_Argv[]);
static int MZ_Proxy(const void *s_ServerAddress, int s_ServerPort, const void *s_ClientAddress, int s_ClientPort, int s_AcceptCount, int s_MaxConnect);
static void MZ_ProxyLog(const void *s_String, ...);

static int __g_CTRL_C__ = 0;

int main(int s_Argc, char *s_Argv[])
{
 int s_Return, s_Index, s_AcceptCount;
 char *s_ServerAddress, *s_ClientAddress;
 int s_ServerPort, s_ClientPort, s_MaxConnect; 

#if DEF_DEBUG_NODAEMON == (0)
 switch(fork())
 { 
  case (-1): return(-1);
  case 0: 
       setsid();
       chdir("/");
       break;
  default: exit(0); break;
 }
#endif
 
 /* Print logo */
 fprintf(stdout, 
	 "MZ_Proxy v1.0.1 - <mailto:minzkn@infoeq.com> - %s %s\n"
	 "Code by JaeHyuk Cho , Application name : TCP Proxy , http://mzproxy.kldp.net\n"
	 "\n",
	 __DATE__, __TIME__
	);
 
 /* Default setup */
 s_ServerAddress = s_ClientAddress = (char *)0;
 s_ServerPort = (-1), s_ClientPort = 0;
 s_AcceptCount = s_MaxConnect = 0;

 /* Parse command line */ 
 s_Index = 1;
 while(s_Index < s_Argc)
 {
  if((s_Index + 1) < s_Argc && s_Argv[s_Index + 1][0] != '-')
  {	 
   if(strcmp(s_Argv[s_Index], "-sa") == 0)
   { /* Server address */
    s_ServerAddress = s_Argv[++s_Index]; 
   }
   else if(strcmp(s_Argv[s_Index], "-sp") == 0)
   { /* Server port */
    if(sscanf(s_Argv[++s_Index], "%i", &s_ServerPort) != 1)s_ServerPort = (-1);
   }
   else if(strcmp(s_Argv[s_Index], "-ca") == 0)
   { /* Client address */
    s_ClientAddress = s_Argv[++s_Index];
   }
   else if(strcmp(s_Argv[s_Index], "-cp") == 0)
   {
    if(sscanf(s_Argv[++s_Index], "%i", &s_ClientPort) != 1)s_ClientPort = (-1);
   }
   else if(strcmp(s_Argv[s_Index], "-ac") == 0)
   {
    if(sscanf(s_Argv[++s_Index], "%i", &s_AcceptCount) != 1)s_AcceptCount = (-1);
   }
   else if(strcmp(s_Argv[s_Index], "-mc") == 0)
   {
    if(sscanf(s_Argv[++s_Index], "%i", &s_MaxConnect) != 1)s_MaxConnect = (-1);
   }
   else break;
  } 
  s_Index++; 
 }

 /* Request check */
 if(s_ServerAddress == (char *)0 || 
    s_ServerPort <= 0 || 
    s_ClientPort <= 0 || 
    s_Index < s_Argc || 
    s_AcceptCount < 0 ||
    s_MaxConnect < 0
   )
 {
  fprintf(stdout, "Option: -sa <ServerAddress>\n"
         	  "        -sp <ServerPort>\n"
          	  "        [-ca <ClientAddress>]\n"
        	  "        -cp <ClientPort>\n"
        	  "        [-ac <AcceptCount>]\n"
		  "        [-mc <MaxConnection>]\n" 
         	  "\n"
        	 );
  if(s_Index < s_Argc)fprintf(stderr, "%s: Invalid option \"%s\" !\n", s_Argv[0], s_Argv[s_Index]);
  s_Return = (-1);
 }
 else 
 {
  fprintf(stdout, "Server address : \"%s\"\n", s_ServerAddress);	 
  fprintf(stdout, "Server port    : \"%d\"\n", s_ServerPort);	 
  fprintf(stdout, "Client address : \"%s\"\n", s_ClientAddress);	 
  fprintf(stdout, "Client port    : \"%d\"\n", s_ClientPort);
  if(s_MaxConnect == 0)fprintf(stdout, "Max connect    : \"%s\"\n", "Unlimited"); 
  else fprintf(stdout, "Max connect    : \"%d\"\n", s_MaxConnect); 
  s_Return = MZ_Proxy(s_ServerAddress, s_ServerPort, s_ClientAddress, s_ClientPort, s_AcceptCount, s_MaxConnect);
 }

 /* Check error print */
 if(s_Return == 1)fprintf(stdout, "Quit MZ_Proxy server.\n");
 else fprintf(stderr, "Quit MZ_Proxy server. (Return code = %d) !\n", s_Return); 
 
 return(s_Return);
}

static void __MZ_CtrlSignal__(int s_Signal)
{
 fprintf(stdout, "CTRL-C break.\n");
 __g_CTRL_C__ = 1;
}

static int MZ_Proxy(const void *s_ServerAddress, int s_ServerPort, const void *s_ClientAddress, int s_ClientPort, int s_AcceptCount, int s_MaxConnect)
{
 int s_Return = 0, s_Socket, s_AcceptIndex, s_IsSelect, s_MaxFD, s_AcceptSocket, s_ConnectSocket, s_RecvBytes, s_SendBytes;
 int s_REUSEADDR_ON = 1, s_ConnectCount = 0;
 struct hostent *s_HostEntry;
 struct sockaddr_in s_BindAddress, s_ConnectAddress, s_AcceptAddress;
 socklen_t s_BindAddressLength, s_ConnectAddressLength, s_AcceptAddressLength;
 fd_set s_FD;
 t_MZ_Proxy *s_ProxyTable, *s_TraceProxyTable;
 unsigned char s_Buffer[ DEF_DEFAULT_BUFFER ];

 /* Setup s_BindAddress */ 
 s_BindAddress.sin_family = AF_INET;
 if(s_ClientAddress)
 {
  s_HostEntry = gethostbyname(s_ClientAddress);
  if(s_HostEntry)memcpy((void *)(&s_BindAddress.sin_addr.s_addr), (void *)(s_HostEntry->h_addr_list[0]), s_HostEntry->h_length);
  else s_Return = (-2);
 }
 else s_BindAddress.sin_addr.s_addr = htonl(INADDR_ANY);
 s_BindAddress.sin_port = htons(s_ClientPort);
 s_BindAddressLength = sizeof(s_BindAddress);

 /* Setup s_ConnectAddress */
 if(s_Return >= 0)
 {
  s_ConnectAddress.sin_family = AF_INET;
  if(s_ServerAddress)
  {
   s_HostEntry = gethostbyname(s_ServerAddress);
   if(s_HostEntry)memcpy((void *)(&s_ConnectAddress.sin_addr.s_addr), (void *)(s_HostEntry->h_addr_list[0]), s_HostEntry->h_length);
   else s_Return = (-2);
  }
  else s_ConnectAddress.sin_addr.s_addr = htonl(INADDR_ANY);
  s_ConnectAddress.sin_port = htons(s_ServerPort);
 }
 s_ConnectAddressLength = sizeof(s_ConnectAddress);

 signal(SIGINT, __MZ_CtrlSignal__);
 
 if(s_Return >= 0)
 {
  s_MaxFD = 0;	    
  s_AcceptIndex = 0;
  s_ProxyTable = (t_MZ_Proxy *)0;
  FD_ZERO(&s_FD); 	     
L_ReFuse_Connect:; 
  s_Socket = socket(AF_INET, SOCK_STREAM, 0);
  if(s_Socket >= 0)
  {
   setsockopt(s_Socket, SOL_SOCKET, SO_REUSEADDR, &s_REUSEADDR_ON, sizeof(s_REUSEADDR_ON));
   if(s_Socket >= s_MaxFD)
   {
    s_MaxFD = s_Socket + 1;	  
#if DEF_DEBUG_PROXY >= (1)
    fprintf(stdout, "Change MaxFD (%d).\n", s_MaxFD);
#endif
   }
   if(bind(s_Socket, (struct sockaddr *)(&s_BindAddress), s_BindAddressLength) == 0)
   {
    if(listen(s_Socket, DEF_DEFAULT_BACKLOG) == 0)
    {
     while((s_AcceptCount == 0 || s_AcceptIndex < s_AcceptCount) && __g_CTRL_C__ == 0)
     {
#if DEF_DEBUG_PROXY >= (2)
      fprintf(stdout, "Ready select (MaxFD = %d).\n", s_MaxFD);
#endif
      /* FD_SET All client */ 
      FD_SET(s_Socket, &s_FD);	   
      for(s_TraceProxyTable = s_ProxyTable;s_TraceProxyTable;s_TraceProxyTable = s_TraceProxyTable->Next)
      {
       if(s_TraceProxyTable->ClientSocket >= 0)FD_SET(s_TraceProxyTable->ClientSocket, &s_FD);
       if(s_TraceProxyTable->ServerSocket >= 0)FD_SET(s_TraceProxyTable->ServerSocket, &s_FD);
      }
      s_IsSelect = select(s_MaxFD, (fd_set *)(&s_FD), (fd_set *)0, (fd_set *)0, (struct timeval *)0);  
      if(s_IsSelect > 0)
      {
#if DEF_DEBUG_PROXY >= (3)
       fprintf(stdout, "Selected.\n");
#endif
       if(FD_ISSET(s_Socket, &s_FD))
       { /* New accept */
	if(s_MaxConnect == 0 || s_ConnectCount < s_MaxConnect)s_ConnectSocket = socket(AF_INET, SOCK_STREAM, 0);
	else
	{
         MZ_ProxyLog("Max connection %d over\n", s_ConnectCount);
         s_ConnectSocket = (-1);
	}
	if(s_ConnectSocket >= 0)
	{
         /* Skip: Not used bind */		
         if(connect(s_ConnectSocket, (struct sockaddr *)(&s_ConnectAddress), s_ConnectAddressLength) == 0)
	 {
#if DEF_DEBUG_PROXY >= (3)
          fprintf(stdout, "To connect.\n");
#endif
	 }
	 else 
	 {
#if defined(WIN32)
	  closesocket(s_ConnectSocket);
	  closesocket(s_Socket);
#else
	  close(s_ConnectSocket);
	  close(s_Socket);
#endif		 
#if DEF_DEBUG_PROXY >= (1)
          fprintf(stdout, "Connect refused (Can not connect to server).\n");
#endif
          MZ_ProxyLog("Can not connect to server (%s:%d)\n", s_ServerAddress, s_ServerPort);
          goto L_ReFuse_Connect;
	 }	 
	}
	else
	{
#if DEF_DEBUG_PROXY >= (1)
         fprintf(stdout, "Connect refused (Can not open socket).\n");
#endif
	 goto L_ReFuse_Connect;
	}
	s_AcceptAddressLength = sizeof(s_AcceptAddress);
	s_AcceptSocket = accept(s_Socket, (struct sockaddr *)(&s_AcceptAddress), (socklen_t *)(&s_AcceptAddressLength));
	if(s_AcceptSocket >= 0)
	{
#if DEF_DEBUG_PROXY >= (1)
         fprintf(stdout, "Accepted from %s:%d.\n", inet_ntoa(s_AcceptAddress.sin_addr), ntohs(s_AcceptAddress.sin_port));
#endif
         if(s_ProxyTable)
	 {
          s_TraceProxyTable = s_ProxyTable;
          while(s_TraceProxyTable->Next)
	  {
           if(s_TraceProxyTable->ClientSocket < 0 && s_TraceProxyTable->ServerSocket < 0)break; /* Reuse proxy table */
           s_TraceProxyTable = s_TraceProxyTable->Next;
	  }	  
	  if(s_TraceProxyTable->ClientSocket >= 0)
	  {
	   s_TraceProxyTable->Next = (t_MZ_Proxy *)malloc(sizeof(t_MZ_Proxy));
           s_TraceProxyTable = s_TraceProxyTable->Next;
	   if(s_TraceProxyTable)s_TraceProxyTable->Next = (t_MZ_Proxy *)0;
	  }
	 }
	 else 
	 {
	  s_TraceProxyTable = s_ProxyTable = (t_MZ_Proxy *)malloc(sizeof(t_MZ_Proxy));
	  if(s_TraceProxyTable)s_TraceProxyTable->Next = (t_MZ_Proxy *)0;
	 }
         if(s_TraceProxyTable)
	 {
          memcpy((void *)(&s_TraceProxyTable->AcceptAddress), (void *)(&s_AcceptAddress), sizeof(s_AcceptAddress));
	  s_TraceProxyTable->AcceptAddressLength = s_AcceptAddressLength;
	  s_TraceProxyTable->ClientSocket = s_AcceptSocket;
	  s_TraceProxyTable->ServerSocket = s_ConnectSocket;
	  if((s_AcceptSocket > s_ConnectSocket ? s_AcceptSocket : s_ConnectSocket) >= s_MaxFD)
	  {
	   s_MaxFD = (s_AcceptSocket > s_ConnectSocket ? s_AcceptSocket : s_ConnectSocket) + 1;
#if DEF_DEBUG_PROXY >= (1)
           fprintf(stdout, "Change MaxFD (%d).\n", s_MaxFD);
#endif
	  }
	  if(s_AcceptCount != 0)s_AcceptIndex++;
          s_TraceProxyTable->FromServerBytes = s_TraceProxyTable->FromClientBytes = (unsigned long long)0;
	  s_ConnectCount++;
#if DEF_DEBUG_PROXY >= (3)
          fprintf(stdout, "Add proxy table.\n");
#endif
	 }
	 else s_Return = (-5);
	}
	else 
	{
#if defined(WIN32)
	 closesocket(s_ConnectSocket);
#else
	 close(s_ConnectSocket);
#endif
#if DEF_DEBUG_PROXY >= (1)
         fprintf(stdout, "Accept failed.\n");
#endif
	 s_Return = (-4);
	}
       }
       else
       { /* Process job */
        for(s_TraceProxyTable = s_ProxyTable;s_TraceProxyTable;s_TraceProxyTable = s_TraceProxyTable->Next)
        {
#if DEF_DEBUG_PROXY >= (4)
          fprintf(stdout, "Check FD (Server=%d, Client=%d)\n", s_TraceProxyTable->ServerSocket, s_TraceProxyTable->ClientSocket);
#endif
         if(s_TraceProxyTable->ClientSocket >= 0)
	 {
          if(FD_ISSET(s_TraceProxyTable->ClientSocket, &s_FD))
	  {
           s_RecvBytes = recv(s_TraceProxyTable->ClientSocket, (void *)(&s_Buffer[0]), sizeof(s_Buffer), MSG_NOSIGNAL);         
	   if(s_RecvBytes > 0)
	   {
            s_SendBytes = send(s_TraceProxyTable->ServerSocket, (void *)(&s_Buffer[0]), s_RecvBytes, MSG_NOSIGNAL);
#if DEF_DEBUG_PROXY >= (1)
            fprintf(stdout, "Client -> Server (%d->%d bytes).\n", s_RecvBytes, s_SendBytes);
#endif
	    if(s_SendBytes == s_RecvBytes)
	    {
	     s_TraceProxyTable->FromClientBytes += s_RecvBytes;
	    }
	    else s_RecvBytes = (-1);
	   }
           if(s_RecvBytes <= 0)
	   { /* Disconnect */
#if defined(WIN32)
	    closesocket(s_TraceProxyTable->ClientSocket), closesocket(s_TraceProxyTable->ServerSocket);
#else
	    close(s_TraceProxyTable->ClientSocket), close(s_TraceProxyTable->ServerSocket);
#endif	  
	    FD_CLR(s_TraceProxyTable->ClientSocket, &s_FD); FD_CLR(s_TraceProxyTable->ServerSocket, &s_FD); 
	    s_TraceProxyTable->ClientSocket = s_TraceProxyTable->ServerSocket = (-1);
#if DEF_DEBUG_PROXY >= (1)
            fprintf(stdout, "Disconnected from client (%s:%d) - %llu(S:%llu, C:%llu) bytes.\n", 
		    inet_ntoa(s_AcceptAddress.sin_addr), 
		    ntohs(s_AcceptAddress.sin_port), 
		    s_TraceProxyTable->FromServerBytes + s_TraceProxyTable->FromClientBytes,
		    s_TraceProxyTable->FromServerBytes,
		    s_TraceProxyTable->FromClientBytes
		   );
#endif
            MZ_ProxyLog("<C> S://%s:%d(%llu) - C://%s:%d(%llu) = %llu.\n",
		    s_ServerAddress, 
		    s_ServerPort, 
		    s_TraceProxyTable->FromServerBytes,
		    inet_ntoa(s_AcceptAddress.sin_addr), 
		    ntohs(s_AcceptAddress.sin_port), 
		    s_TraceProxyTable->FromClientBytes,
		    s_TraceProxyTable->FromServerBytes + s_TraceProxyTable->FromClientBytes
		   );
	    s_ConnectCount--;
	   }
	  }
	 }
	 if(s_TraceProxyTable->ServerSocket >= 0)
	 {
          if(FD_ISSET(s_TraceProxyTable->ServerSocket, &s_FD))
	  {
           s_RecvBytes = recv(s_TraceProxyTable->ServerSocket, (void *)(&s_Buffer[0]), sizeof(s_Buffer), MSG_NOSIGNAL);         
	   if(s_RecvBytes > 0)
	   {
            s_SendBytes = send(s_TraceProxyTable->ClientSocket, (void *)(&s_Buffer[0]), s_RecvBytes, MSG_NOSIGNAL);
#if DEF_DEBUG_PROXY >= (1)
            fprintf(stdout, "Server -> Client (%d->%d bytes).\n", s_RecvBytes, s_SendBytes);
#endif
	    if(s_SendBytes == s_RecvBytes)
	    {
	     s_TraceProxyTable->FromServerBytes += s_RecvBytes;
	    }
	    else s_RecvBytes = (-1);
	   }
           if(s_RecvBytes <= 0)
	   { /* Disconnect */
#if defined(WIN32)
	    closesocket(s_TraceProxyTable->ClientSocket), closesocket(s_TraceProxyTable->ServerSocket);
#else
	    close(s_TraceProxyTable->ClientSocket), close(s_TraceProxyTable->ServerSocket);
#endif	  
	    FD_CLR(s_TraceProxyTable->ClientSocket, &s_FD); FD_CLR(s_TraceProxyTable->ServerSocket, &s_FD); 
	    s_TraceProxyTable->ClientSocket = s_TraceProxyTable->ServerSocket = (-1);
#if DEF_DEBUG_PROXY >= (1)
            fprintf(stdout, "Disconnected from server (%s:%d) - %llu(S:%llu, C:%llu) bytes.\n", 
		    inet_ntoa(s_AcceptAddress.sin_addr), 
		    ntohs(s_AcceptAddress.sin_port), 
		    s_TraceProxyTable->FromServerBytes + s_TraceProxyTable->FromClientBytes,
		    s_TraceProxyTable->FromServerBytes,
		    s_TraceProxyTable->FromClientBytes
		   );
#endif
            MZ_ProxyLog("<S> S://%s:%d(%llu) - C://%s:%d(%llu) = %llu.\n",
		    s_ServerAddress, 
		    s_ServerPort, 
		    s_TraceProxyTable->FromServerBytes,
		    inet_ntoa(s_AcceptAddress.sin_addr), 
		    ntohs(s_AcceptAddress.sin_port), 
		    s_TraceProxyTable->FromClientBytes,
		    s_TraceProxyTable->FromServerBytes + s_TraceProxyTable->FromClientBytes
		   );
	    s_ConnectCount--;
	   }
	  }
	 }
	}
       }
      } /* s_IsSelect > 0 */
      /* else s_Return = (-5); */
     }
    }
    else s_Return = (-4);
   } 
   else s_Return = (-3);
#if defined(WIN32)
   closesocket(s_Socket);
#else
   close(s_Socket);
#endif
  }
 
  /* Free proxy table */ 
  while(s_ProxyTable)
  {
   s_TraceProxyTable = s_ProxyTable;	  
   s_ProxyTable = s_ProxyTable->Next;
   if(s_TraceProxyTable->ServerSocket >= 0)
   {
#if defined(WIN32)
    closesocket(s_TraceProxyTable->ServerSocket);
#else
    close(s_TraceProxyTable->ServerSocket);
#endif
   }
   if(s_TraceProxyTable->ClientSocket >= 0)
   {
#if defined(WIN32)
    closesocket(s_TraceProxyTable->ClientSocket);
#else
    close(s_TraceProxyTable->ClientSocket);
#endif
   }
   free(s_TraceProxyTable);
  } 
 }
 
 return(s_Return);
}

static void MZ_ProxyLog(const void *s_String, ...)
{
 void *s_Message;
 int s_BufferSize = 256, s_Handle;
 time_t s_Time_t;
 struct tm *s_TM;
 char s_TimeStampString[ 64 ];
 va_list s_VaArg;
 s_Message = DEF_DEFAULT_LOG;
 if(s_Message)
 {
  time(&s_Time_t);
  s_TM = localtime(&s_Time_t);
  sprintf(s_TimeStampString, "[%d/%d/%d %02d:%02d:%02d] ",
          s_TM->tm_year + 1900, s_TM->tm_mon + 1, s_TM->tm_mday,
          s_TM->tm_hour, s_TM->tm_min, s_TM->tm_sec
	 );
  va_start(s_VaArg, s_String);
  do
  {
   s_Message = (void *)malloc(s_BufferSize);
   if(s_Message)
   {
    /* vsnprintf ! */
    if(vsnprintf(s_Message, s_BufferSize - 1, s_String, s_VaArg) < s_BufferSize)break; 
    else free(s_Message); 
   }
   else break;
   s_Message = (void *)0;
   s_BufferSize += 256;
  }while(s_BufferSize < 8192);
  va_end(s_VaArg);
  if(s_Message)
  {
   s_Handle = open(DEF_DEFAULT_LOG, O_RDWR | O_CREAT | O_SYNC, 0644); 
   if(s_Handle >= 0)
   {
    if(lseek(s_Handle, 0, SEEK_END) > (1 << 20))
    {
     lseek(s_Handle, 0, SEEK_SET);
     ftruncate(s_Handle, 0);
    }	    
    write(s_Handle, s_TimeStampString, strlen(s_TimeStampString));
    write(s_Handle, s_Message, strlen(s_Message));
    close(s_Handle);
   }
   free(s_Message);
  } 
 }
}

/* End of source */
